/*
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 * Copyright (C) 2012-2013 RIVIÈRE Aurélien <aurelien.riv@gmail.com>
 *
 */

#include "main.h"

#define LWP_CONFIG
#include "lw-gui-common.h"


#define LW_CONFIG_APPLICATION_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_CONFIG_TYPE_APPLICATION, LwConfigApplicationPrivate))

struct _LwConfigApplicationPrivate
{
	GtkListStore *store;

	GtkWidget *content;
	GtkWidget *plugin_icon;
	GtkWidget *plugin_name;
	GtkWidget *plugin_switch;
	GtkWidget *plugin_desc;

	LwConfigPlugin *selected;

	GSettings *settings;
	gchar *active_plugin;
};

enum
{
	PROP_0,

	PROP_ACTIVE_PLUGIN,

	N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(LwConfigApplication, lw_config_application, GTK_TYPE_APPLICATION)


LwConfigApplication*
lw_config_application_new()
{
	return g_object_new(LW_CONFIG_TYPE_APPLICATION,
	                    "application-id", GSETTINGS_PREFIX"livewallpaper-config",
	                    "flags", G_APPLICATION_FLAGS_NONE,
	                    NULL);
}

static void
lw_config_application_set_property(GObject *object,
                                   guint property_id,
                                   const GValue *value,
                                   GParamSpec *pspec)
{
	LwConfigApplication *self = LW_CONFIG_APPLICATION(object);

	switch(property_id)
	{
		case PROP_ACTIVE_PLUGIN:
			g_free(self->priv->active_plugin);
			self->priv->active_plugin = g_strdup(g_value_get_string(value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
lw_config_application_get_property(GObject *object,
                                   guint property_id,
                                   GValue *value,
                                   GParamSpec *pspec)
{
	LwConfigApplication *self = LW_CONFIG_APPLICATION(object);

	switch(property_id)
	{
		case PROP_ACTIVE_PLUGIN:
			g_value_set_string(value, self->priv->active_plugin);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}


static gboolean
plugin_select_func(GtkTreeSelection *selection,
                   GtkTreeModel *model,
                   GtkTreePath *path,
                   gboolean selected,
                   gpointer data)
{
	GtkTreeIter iter;
	gboolean is_plugin;

	(void) selection;
	(void) selected;
	(void) data;

	gtk_tree_model_get_iter(model, &iter, path);
	gtk_tree_model_get(model, &iter, PLUGIN_ROW_COL, &is_plugin, -1);

	return is_plugin;
}

static void
selected_plugin_changed_cb(GtkTreeSelection *selection, LwConfigApplication *app)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if(gtk_tree_selection_get_selected(selection, &model, &iter))
	{
		LwConfigPlugin *p;
		gchar *markup;

		if(app->priv->selected != NULL)
		{
			/* Remove current settings widget */
			gtk_container_remove(GTK_CONTAINER(app->priv->content),
			                     lw_config_plugin_get_settings_widget(app->priv->selected));
		}

		gtk_tree_model_get(model, &iter,
		                   PLUGIN_COL, &p,
		                   -1);

		/* Icon */
		gtk_image_set_from_pixbuf(GTK_IMAGE(app->priv->plugin_icon),
		                          lw_config_plugin_get_icon(p, 64));

		/* Name */
		markup = g_strconcat("<span weight='bold' size='x-large'>",
		                     lw_config_plugin_get_name(p), "</span>", NULL);
		gtk_label_set_markup(GTK_LABEL(app->priv->plugin_name), markup);
		g_free(markup);

		/* Switch */
		if(strcmp(lw_config_plugin_get_id(p), "general") == 0)
			gtk_widget_show(app->priv->plugin_switch);
		else
			gtk_widget_hide(app->priv->plugin_switch);

		/* Description */
		gtk_label_set_text(GTK_LABEL(app->priv->plugin_desc),
		                   lw_config_plugin_get_description(p));

		/* Add settings widget */
		gtk_box_pack_end(GTK_BOX(app->priv->content),
		                 lw_config_plugin_get_settings_widget(p),
		                 TRUE, TRUE, 0);
		gtk_widget_show_all(app->priv->content);

		app->priv->selected = p;
	}
}

static void
is_active_plugin_func(GtkCellLayout *layout,
                      GtkCellRenderer *cell,
                      GtkTreeModel *model,
                      GtkTreeIter *iter,
                      LwConfigApplication *app)
{
	LwConfigPlugin *p;
	gboolean is_active = FALSE;

	(void) layout;

	gtk_tree_model_get(model, iter,
	                   PLUGIN_COL, &p,
	                   -1);

	if(p == NULL) return;

	if(g_strcmp0(lw_config_plugin_get_id(p), app->priv->active_plugin) == 0)
		is_active = TRUE;

	g_object_set(G_OBJECT(cell), "active", is_active, NULL);
}

static void
activate_plugin_cb(GtkCellRenderer *cell,
                   gchar *path_str,
                   LwConfigApplication *app)
{
	LwConfigPlugin *p;
	GtkTreeIter iter;
	GtkTreePath *path = gtk_tree_path_new_from_string(path_str);
	GtkTreeModel *model = GTK_TREE_MODEL(app->priv->store);

	(void) cell;

	gtk_tree_model_get_iter(model, &iter, path);
	gtk_tree_model_get(model, &iter,
	                   PLUGIN_COL, &p,
	                   -1);

	g_object_set(G_OBJECT(app), "active-plugin", lw_config_plugin_get_id(p), NULL);

	/* Notify the GtkTreeView about this change */
	gtk_tree_model_row_changed(model, path, &iter);

	gtk_tree_path_free(path);
}

static void
show_plugin_info_cb(GtkButton *button, LwConfigApplication *app)
{
	GtkAboutDialog *dialog = lw_config_plugin_create_about_dialog(app->priv->selected);

	(void) button;

	switch(gtk_dialog_run(GTK_DIALOG(dialog)))
	{
		case GTK_RESPONSE_CANCEL:
		case GTK_RESPONSE_DELETE_EVENT:
			gtk_widget_destroy(GTK_WIDGET(dialog));
			break;
	}
}


static void
lw_config_application_init_plugins(LwConfigApplication *self)
{
	LwConfigPlugin *plugin;
	GSList *plugins = NULL;
	gchar *filename, *module_dir, *data_dir;

	/* Add general settings */
	filename = g_build_filename(LIBDIR, "livewallpaper", "livewallpaper.plugin", NULL);
	plugin = lw_config_plugin_new(filename, LIBDIR, DATADIR);
	g_free(filename);

	if(plugin != NULL)
		lw_config_plugin_store_add_plugin(self->priv->store, plugin,
		                                  GENERAL_SETTINGS_POS, FALSE);

	/* Find local plugins */
	module_dir = g_build_filename(g_get_home_dir(), ".livewallpaper", "plugins", NULL);
	plugins = lw_config_find_plugins(module_dir, module_dir);
	g_free(module_dir);

	lw_config_plugin_store_add_plugins(self->priv->store, plugins);

	/* Find plugins in the install directory */
	module_dir = g_build_filename(LIBDIR, "livewallpaper", "plugins", NULL);
	data_dir = g_build_filename(DATADIR, "livewallpaper", "plugins", NULL);
	plugins = lw_config_find_plugins(module_dir, data_dir);
	g_free(module_dir);
	g_free(data_dir);

	lw_config_plugin_store_add_plugins(self->priv->store, plugins);
}

static void
lw_config_application_create_window(LwConfigApplication *self)
{
	GtkWindow *window;
	GtkWidget *hbox, *vbox;

	GtkWidget *scrolled_window, *tree_view;
	GtkTreeViewColumn *column;
	GtkCellRenderer *cell;
	GtkTreeSelection *selection;

	GtkWidget *grid, *tmp, *button_box;

#if GTK_CHECK_VERSION(3,4,0)
	window = GTK_WINDOW(gtk_application_window_new(GTK_APPLICATION(self)));
#else
	window = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
	gtk_window_set_application(window, GTK_APPLICATION(self));
#endif

	gtk_window_set_title        (window, _("LiveWallpaper Preferences"));
	gtk_window_set_icon_name    (window, "livewallpaper-config");
	gtk_window_set_default_size (window, 900, 500);
	gtk_container_set_border_width(GTK_CONTAINER(window), 12);

	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 12);
	gtk_container_add(GTK_CONTAINER(window), hbox);


	/* Left side */
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, TRUE, 0);

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
	                               GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);

	/* PluginStore */
	self->priv->store = lw_config_plugin_store_new();

	/* TreeView */
	tree_view = gtk_tree_view_new();
	gtk_container_add(GTK_CONTAINER(scrolled_window), tree_view);

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), FALSE);
	gtk_tree_view_set_model(GTK_TREE_VIEW(tree_view), GTK_TREE_MODEL(self->priv->store));

	/* Columns */
	column = gtk_tree_view_column_new();

	cell = gtk_cell_renderer_pixbuf_new();
	gtk_cell_layout_pack_start    (GTK_CELL_LAYOUT(column), cell, FALSE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "pixbuf", ICON_COL);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "visible", PLUGIN_ROW_COL);
	gtk_cell_renderer_set_padding (GTK_CELL_RENDERER(cell), 6, 0);

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start    (GTK_CELL_LAYOUT(column), cell, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "markup", NAME_COL);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "visible", PLUGIN_ROW_COL);
	gtk_cell_renderer_set_padding (GTK_CELL_RENDERER(cell), 6, 0);

	cell = gtk_cell_renderer_toggle_new();
	gtk_cell_layout_pack_start         (GTK_CELL_LAYOUT(column), cell, FALSE);
	gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT(column), cell,
	                                    (GtkCellLayoutDataFunc) is_active_plugin_func, self, NULL);
	gtk_cell_layout_add_attribute      (GTK_CELL_LAYOUT(column), cell, "visible", TOGGLE_ROW_COL);
	gtk_cell_renderer_toggle_set_radio (GTK_CELL_RENDERER_TOGGLE(cell), TRUE);
	gtk_cell_renderer_set_padding      (GTK_CELL_RENDERER(cell), 12, 0);
	g_signal_connect(cell, "toggled", G_CALLBACK(activate_plugin_cb), self);

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start    (GTK_CELL_LAYOUT(column), cell, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "text", TITLE_COL);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "visible", HEADING_ROW_COL);
	gtk_cell_renderer_set_padding (GTK_CELL_RENDERER(cell), 6, 6);
	g_object_set(G_OBJECT(cell),
	             "weight", PANGO_WEIGHT_BOLD,
	             "weight-set", TRUE,
	             NULL);

	gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), column);

	/* Selection */
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
	gtk_tree_selection_set_select_function(selection, plugin_select_func, NULL, NULL);
	g_signal_connect(selection, "changed", G_CALLBACK(selected_plugin_changed_cb), self);


	/* Right side */
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 12);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
	self->priv->content = vbox;

	/* Plugin Information */
	grid = gtk_grid_new();
	gtk_box_pack_start(GTK_BOX(vbox), grid, FALSE, TRUE, 0);

	tmp = gtk_image_new();
	gtk_misc_set_padding(GTK_MISC(tmp), 6, 6);
	gtk_grid_attach(GTK_GRID(grid), tmp, 0, 0, 2, 2);
	self->priv->plugin_icon = tmp;

	tmp = gtk_label_new("Name");
	gtk_widget_set_hexpand(tmp, TRUE);
	gtk_widget_set_halign(tmp, GTK_ALIGN_START);
	gtk_widget_set_valign(tmp, GTK_ALIGN_END);
	gtk_misc_set_padding(GTK_MISC(tmp), 6, 2);
	gtk_grid_attach(GTK_GRID(grid), tmp, 2, 0, 3, 1);
	self->priv->plugin_name = tmp;

	button_box = gtk_button_box_new(GTK_ORIENTATION_HORIZONTAL);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box), GTK_BUTTONBOX_END);
	gtk_widget_set_valign(button_box, GTK_ALIGN_END);
	gtk_grid_attach(GTK_GRID(grid), button_box, 5, 0, 1, 1);

	tmp = gtk_switch_new();
	gtk_widget_set_no_show_all(tmp, TRUE);
	gtk_widget_set_tooltip_text(tmp, _("Show or hide the live wallpaper. This will start LiveWallpaper if it is not already running."));
	gtk_box_pack_start(GTK_BOX(button_box), tmp, TRUE, TRUE, 0);

	g_settings_bind(self->priv->settings, "active", tmp, "active", G_SETTINGS_BIND_DEFAULT);
	g_signal_connect(G_OBJECT(tmp), "activate", G_CALLBACK(livewallpaper_button_activated),
	                 self->priv->settings);
	g_bus_watch_name(G_BUS_TYPE_SESSION, "net.launchpad.LiveWallpaper", G_BUS_NAME_WATCHER_FLAGS_NONE,
	                 (GBusNameAppearedCallback) core_appeared, (GBusNameVanishedCallback) core_vanished,
	                 tmp, NULL);
	self->priv->plugin_switch = tmp;

	tmp = gtk_label_new("Description");
	gtk_widget_set_hexpand(tmp, TRUE);
	gtk_widget_set_halign(tmp, GTK_ALIGN_START);
	gtk_widget_set_valign(tmp, GTK_ALIGN_START);
	gtk_misc_set_padding(GTK_MISC(tmp), 6, 2);
	gtk_misc_set_alignment(GTK_MISC(tmp), 0, 1);
	gtk_label_set_line_wrap(GTK_LABEL(tmp), TRUE);
	gtk_grid_attach(GTK_GRID(grid), tmp, 2, 1, 4, 1);
	self->priv->plugin_desc = tmp;

	/* About Button */
	button_box = gtk_button_box_new(GTK_ORIENTATION_HORIZONTAL);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box), GTK_BUTTONBOX_END);
	gtk_box_pack_end(GTK_BOX(vbox), button_box, FALSE, TRUE, 0);

	tmp = gtk_button_new_with_label("Info");
	gtk_box_pack_start(GTK_BOX(button_box), tmp, TRUE, TRUE, 0);
	g_signal_connect(tmp, "clicked", G_CALLBACK(show_plugin_info_cb), self);


	/* Find all plugins */
	lw_config_application_init_plugins(self);

	/* Show window */
	gtk_widget_show_all(GTK_WIDGET(window));
}

static void
lw_config_application_activate(GApplication *app)
{
	LwConfigApplication *self = LW_CONFIG_APPLICATION(app);
    GList *list = gtk_application_get_windows(GTK_APPLICATION(app));

    if(list)	/* If an instance of the app is running, we put its window to the foreground */
		gtk_window_present(GTK_WINDOW(list->data));
    else		/* Else we create the window */
    	lw_config_application_create_window(self);
}

static void
lw_config_application_init(LwConfigApplication *self)
{
	self->priv = LW_CONFIG_APPLICATION_GET_PRIVATE(self);

	self->priv->settings = g_settings_new("net.launchpad.livewallpaper");
	self->priv->active_plugin = NULL;

	g_settings_bind(self->priv->settings, "active-plugin", self, "active-plugin", G_SETTINGS_BIND_DEFAULT);
}

static void
lw_config_application_dispose(GObject *object)
{
	LwConfigApplication *self = LW_CONFIG_APPLICATION(object);

	if(self->priv->settings)
	{
		g_object_unref(self->priv->settings);
		self->priv->settings = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_config_application_parent_class)->dispose(object);
}

static void
lw_config_application_finalize(GObject *object)
{
	LwConfigApplication *self = LW_CONFIG_APPLICATION(object);

	g_free(self->priv->active_plugin);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_config_application_parent_class)->finalize(object);
}

static void
lw_config_application_class_init(LwConfigApplicationClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GApplicationClass *gapplication_class = G_APPLICATION_CLASS(klass);

	gobject_class->set_property = lw_config_application_set_property;
	gobject_class->get_property = lw_config_application_get_property;
	gobject_class->dispose = lw_config_application_dispose;
	gobject_class->finalize = lw_config_application_finalize;

	gapplication_class->activate = lw_config_application_activate;

	g_type_class_add_private(klass, sizeof(LwConfigApplicationPrivate));

	obj_properties[PROP_ACTIVE_PLUGIN] =
		g_param_spec_string("active-plugin",
		                    "Active Plugin",
		                    "ID of the active plugin",
		                    NULL,
		                    G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

