/*
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 RIVIÈRE Aurélien <aurelien.riv@gmail.com>
 *
 */

#include "main.h"
#include "loader-internal.h"


static void files_selector_get_value (GSettings *, gchar *, GtkWidget *);
static void files_selector_set_value (GtkWidget *);
static void files_selector_add_value (GtkWidget *, gpointer);
static void files_selector_rm_value  (GtkWidget *, gpointer);

GtkWidget *
files_selector_new (Key *k, GSettings *settings, gboolean pictures_filter)
{
	gchar *signame = g_strdup_printf("changed::%s", k->name);

	GtkWidget *box     = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0),
			  *toolbar = gtk_toolbar_new (),
			  *scroll  = gtk_scrolled_window_new (NULL, NULL);
	GtkToolItem *ctl;
	GtkStyleContext *context;

	GtkTreeViewColumn *column;
	GtkCellRenderer *cell;

	/* Tree View */
	GtkWidget *treeView = gtk_tree_view_new ();

	column = gtk_tree_view_column_new();
	if(pictures_filter)
		gtk_tree_view_column_set_title(column, _("Pictures"));
	else
		gtk_tree_view_column_set_title(column, _("Folders"));

	cell = gtk_cell_renderer_pixbuf_new();
	gtk_cell_layout_pack_start    (GTK_CELL_LAYOUT(column), cell, FALSE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "pixbuf", 0);

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start    (GTK_CELL_LAYOUT(column), cell, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT(column), cell, "markup", 1);

	gtk_tree_view_append_column(GTK_TREE_VIEW(treeView), column);

	g_object_set_data (G_OBJECT(treeView), "settings", settings);
	g_object_set_data (G_OBJECT(treeView), "key", g_strdup(k->name));
	g_signal_connect (settings, signame, G_CALLBACK(files_selector_get_value), treeView);
	gtk_container_add (GTK_CONTAINER(scroll), treeView);
	g_free (signame);

	/* Scrolled Window */
	gtk_widget_set_size_request(scroll, 400, 50);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(scroll), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);
	gtk_box_pack_start (GTK_BOX(box), scroll, TRUE, TRUE, 0);

	files_selector_get_value (settings, k->name, treeView);

	gtk_toolbar_set_icon_size(GTK_TOOLBAR(toolbar), GTK_ICON_SIZE_MENU);

	/* Button DEL */
	ctl = gtk_tool_button_new (NULL, NULL);
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(ctl), "list-remove-symbolic");
	gtk_widget_set_tooltip_text(GTK_WIDGET(ctl), _("Remove selected item"));
	g_signal_connect (ctl, "clicked", G_CALLBACK(files_selector_rm_value), treeView);
	gtk_toolbar_insert (GTK_TOOLBAR(toolbar), ctl, 0);

	/* Button ADD */
	ctl = gtk_tool_button_new (NULL, NULL);
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(ctl), "list-add-symbolic");
	gtk_widget_set_tooltip_text(GTK_WIDGET(ctl), _("Add new item"));
	g_object_set_data (G_OBJECT(ctl), "pictures_filter", GINT_TO_POINTER(pictures_filter));
	g_signal_connect (ctl, "clicked", G_CALLBACK(files_selector_add_value), treeView);
	gtk_toolbar_insert (GTK_TOOLBAR(toolbar), ctl, 0);

	/* Change style context */
	context = gtk_widget_get_style_context (scroll);
	gtk_style_context_set_junction_sides (context, GTK_JUNCTION_BOTTOM);
	context = gtk_widget_get_style_context (toolbar);
	gtk_style_context_set_junction_sides (context, GTK_JUNCTION_TOP);
	gtk_style_context_add_class (context, "inline-toolbar");

	gtk_box_pack_end (GTK_BOX(box), toolbar, FALSE, FALSE, 0);

	return box;
}

static GdkPixbuf*
get_pixbuf_for_path(const gchar *path)
{
	GdkPixbuf *pixbuf = NULL;
	GFile *file = g_file_new_for_path(path);
	GFileInfo *info = g_file_query_info(file, G_FILE_ATTRIBUTE_STANDARD_ICON","G_FILE_ATTRIBUTE_THUMBNAIL_PATH,
	                                    G_FILE_QUERY_INFO_NONE, NULL, NULL);

	const gchar *thumbnail = g_file_info_get_attribute_byte_string(info, G_FILE_ATTRIBUTE_THUMBNAIL_PATH);
	if(thumbnail)
		pixbuf = gdk_pixbuf_new_from_file_at_size(thumbnail, 16, 16, NULL);

	if(pixbuf == NULL)
	{
		GIcon *icon = g_file_info_get_icon(info);
		GtkIconInfo *icon_info = gtk_icon_theme_lookup_by_gicon(gtk_icon_theme_get_default(),
		                                                        icon, 16, GTK_ICON_LOOKUP_FORCE_SIZE);

		if(icon_info != NULL)
		{
			pixbuf = gtk_icon_info_load_icon(icon_info, NULL);
			gtk_icon_info_free(icon_info);
		}
	}

	g_object_unref(info);
	g_object_unref(file);

	return pixbuf;
}

static void
files_selector_get_value (GSettings *s, gchar *key, GtkWidget *w)
{
	gint i;
	gchar **files 		= g_settings_get_strv (s,key);
	GtkListStore *store = gtk_list_store_new (2, GDK_TYPE_PIXBUF, G_TYPE_STRING);

	for (i=0; files[i]; i++)
	{
		GtkTreeIter iter;
		GdkPixbuf *pixbuf = get_pixbuf_for_path(files[i]);

		gtk_list_store_append (store, &iter);
		gtk_list_store_set 	  (store, &iter, 0, pixbuf, 1, files[i], -1);

		g_object_unref(pixbuf);
	}

	gtk_tree_view_set_model (GTK_TREE_VIEW(w), GTK_TREE_MODEL(store));
	g_object_unref (G_OBJECT(store));
}

static void
files_selector_set_value (GtkWidget *w)
{
	gchar **files = NULL;
	GtkTreeIter iter;
	GtkTreeModel *store = gtk_tree_view_get_model (GTK_TREE_VIEW(w));

	/* Initialize iter so that it point to the first element of the list,
	 * or skip if the list is empty */
	if (gtk_tree_model_get_iter_first (store, &iter))
	{
		guint i = 0;

		/* Count the number of elements of the list */
		while (++i && gtk_tree_model_iter_next (store, &iter));

		files = g_malloc( (i+1) * sizeof(gchar *) );
		i = 0;

		/* Reset the iter to the first element and add each value to file[] */
		gtk_tree_model_get_iter_first (store, &iter);
		do
			gtk_tree_model_get (store, &iter, 1, &files[i], -1);
		while (++i && gtk_tree_model_iter_next (store, &iter));

		files[i] = NULL;
	}
	
	g_settings_set_strv (g_object_get_data (G_OBJECT(w), "settings"),
						 g_object_get_data (G_OBJECT(w), "key"),
						 (const gchar * const*) files);
	g_free (files);
}

static void
files_selector_add_value (GtkWidget *w, gpointer p)
{
	gboolean pictures_filter = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(w), "pictures_filter"));
	
	/* Create a GtkFileChooserDialog with two buttons Open and Cancel */
	GtkWidget *dialog = gtk_file_chooser_dialog_new (NULL, NULL, 
													 pictures_filter ? GTK_FILE_CHOOSER_ACTION_OPEN : GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
													 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
													 GTK_STOCK_OPEN  , GTK_RESPONSE_ACCEPT, NULL);
	gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER(dialog), TRUE);
	
	if (pictures_filter)
	{
		GtkFileFilter *filter = gtk_file_filter_new ();
		gtk_file_filter_add_pixbuf_formats (filter);
		gtk_file_chooser_set_filter ( GTK_FILE_CHOOSER (dialog), filter);
	}
	
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
	{
		GSList *files = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER (dialog));
		GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model (GTK_TREE_VIEW(p)));
		guint i;
		
		for (i = 0; i < g_slist_length (files); i++)
		{
			GtkTreeIter iter;
			gchar *path = g_slist_nth_data (files, i);
			GdkPixbuf *pixbuf = get_pixbuf_for_path(path);

			gtk_list_store_append (store, &iter);
			gtk_list_store_set 	  (store, &iter, 0, pixbuf, 1, path, -1);

			g_object_unref(pixbuf);
		}

		g_slist_free_full (files, g_free);
		files_selector_set_value (p);
	}

	gtk_widget_destroy (dialog);

	(void) w;
}

static void
files_selector_rm_value (GtkWidget *w, gpointer p)
{
	GtkTreeSelection *treeSelection = gtk_tree_view_get_selection (GTK_TREE_VIEW(p));
	if (gtk_tree_selection_count_selected_rows (treeSelection) > 0)
	{
		GtkTreeIter iter;
		GtkTreeModel *store = gtk_tree_view_get_model (GTK_TREE_VIEW(p));

		gtk_tree_selection_get_selected (treeSelection, &store, &iter);
		gtk_list_store_remove (GTK_LIST_STORE(store), &iter);

		files_selector_set_value (p);
		
		(void) w;
	}
}

