/*
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 RIVIÈRE Aurélien <aurelien.riv@gmail.com>
 *
 */

#include "main.h"
#include "loader-internal.h"


#define SPACE 	 	  "%*[ \t]"
#define STR_ATTR(len) "%*c" "%"#len"[^\">]" "%*c"
#define STR_TEXT(len) "%*c" "%"#len"[^\"</\"]"

static guint nb_decimals  (gdouble x);

static GtkWidget *load_enum     (Key *, const gchar *, GSettings *);
static GtkWidget *create_widget (Key *, GSettings *);
static void       add_key       (Key *, GtkWidget *, GSettings *, GtkWidget *);

static void reset_key (GtkWidget *, GSettings *);

static void cb_bind_set_value (GtkWidget *, GSettings *);
static void cb_bind_get_value (GSettings *, gchar *, GtkWidget *);

static gboolean s_to_rgba  (GValue *,       GVariant *,           gpointer);
static GVariant *rgba_to_s (const GValue *, const GVariantType *, gpointer);

static void file_selector_set_value (GtkWidget *, GSettings *);
static void file_selector_get_value (GSettings *, gchar *, GtkWidget *);

static guint
nb_decimals  (gdouble x)
{
	guint n = 7, /* number of decimals */
		  b = 10;
	x = x*10e6;

	while (! ((guint)x % b) && n--)
		b *= 10;

	return n;
}

static GtkWidget*
load_enum (Key *k, const gchar *filename, GSettings *settings)
{
	FILE *file = NULL;
	GtkWidget *ctl = NULL;
	gboolean found = FALSE;

	if(! (file = fopen(filename, "r")))
		perror(filename);
	else
	{
		gchar str[1024];

		/* Read the entire file line by line */
		while (fgets(str, 1024, file))
		{
			if (strstr(str, "<enum"))
			{
				gchar enum_id[128];
				sscanf(strpbrk(str,"<"), "<enum" SPACE "id=" STR_ATTR(127), enum_id);

				if(strcmp(enum_id, k->enum_id) == 0)
				{
					found = TRUE;

					ctl = gtk_combo_box_text_new ();
					gtk_widget_set_size_request(ctl, 200, 0);
					/* Join the key name to the GtkComboBoxText */
					g_object_set_data (G_OBJECT(ctl), "key", g_strdup(k->name));
				}
			}
			else if (found && strstr(str, "<value"))
			{
				gchar nick[64];
				sscanf(strpbrk(str,"n"), "nick=" STR_ATTR(63), nick);

				/*
				 * Append the found value to the GtkComboBoxText with the untranslated nick as
				 * identifier and the translated nick as displayed string
				 */
				gtk_combo_box_text_insert (GTK_COMBO_BOX_TEXT(ctl), -1, nick, _(nick));
			}
			else if (found && strstr(str,"</enum>"))
			{
				/* Bind the "changed" signal to the widget and the "changed::name" to gsettings */
				gchar *signame = g_strdup_printf("changed::%s", k->name);
				g_signal_connect (ctl, "changed", G_CALLBACK(cb_bind_set_value), settings);
				g_signal_connect (settings, signame, G_CALLBACK(cb_bind_get_value), ctl);
				g_free (signame);

				/* Update the widget */
				cb_bind_get_value (settings, k->name, ctl);

				fclose(file);
				return ctl;
			}
			else if (strstr(str, "<lwp:tab"))
				break;
		}

		fclose(file);
	}

	return NULL;
}

static GtkWidget*
create_widget (Key *k, GSettings *settings)
{
	GtkWidget *ctl = NULL;

	switch (k->type)
	{
		case 'b':
			ctl = gtk_switch_new ();
			g_settings_bind (settings, k->name, ctl, "active", G_SETTINGS_BIND_DEFAULT);
			break;

		case 'n': case 'q':		/* int16 uint16 */
		case 'i': case 'u':		/* int   uint   */
		case 'x': case 't':		/* int64 uint64 */
		case 'y': case 'd':		/* byte  double */
			ctl = gtk_spin_button_new_with_range ( k->min, k->max, k->inc);

			if (k->type == 'd')
				gtk_spin_button_set_digits (GTK_SPIN_BUTTON(ctl), k->digits);

			g_settings_bind (settings, k->name, ctl, "value", G_SETTINGS_BIND_DEFAULT);
			gtk_widget_set_size_request(ctl, 100, 0);
			break;

		case 'c':	/* Color */
			ctl = gtk_color_button_new ();
#if GTK_CHECK_VERSION(3,4,0)
			gtk_color_chooser_set_use_alpha (GTK_COLOR_CHOOSER(ctl), TRUE);
#else
			gtk_color_button_set_use_alpha (GTK_COLOR_BUTTON(ctl), TRUE);
#endif
			g_settings_bind_with_mapping (settings, k->name, ctl,
				"rgba", G_SETTINGS_BIND_DEFAULT, s_to_rgba, rgba_to_s, NULL, NULL);
			break;

		case 'p':	/* Picture */
		case 'F':	/* Folder */
		{
			gchar *signame = g_strdup_printf("changed::%s", k->name);

			if (k->type == 'p')
			{
				GtkFileFilter *filter = gtk_file_filter_new ();

				ctl = gtk_file_chooser_button_new (_("Select a picture"), GTK_FILE_CHOOSER_ACTION_OPEN);
				gtk_file_filter_add_pixbuf_formats (filter);
				gtk_file_chooser_set_filter ( GTK_FILE_CHOOSER (ctl), filter);
			}
			else
				ctl = gtk_file_chooser_button_new (_("Select a directory"), GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);

			g_object_set_data (G_OBJECT (ctl), "key", g_strdup(k->name));
			g_signal_connect ( ctl, "file-set", G_CALLBACK(file_selector_set_value), settings);
			g_signal_connect (settings, signame, G_CALLBACK(file_selector_get_value), ctl);
			g_free (signame);

			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (ctl), g_settings_get_string (settings, k->name));
			break;
		}

		case 'P':	/* Pictures */
		case 'D':	/* Directories */
			ctl = files_selector_new (k, settings, k->type == 'P');
			gtk_widget_set_size_request(ctl, 200, 0);
			break;

		case 'f':	/* Font */
			ctl = gtk_font_button_new ();
			gtk_widget_set_size_request(ctl, 200, 0);
			g_settings_bind (settings, k->name, ctl, "font-name", G_SETTINGS_BIND_DEFAULT);
			break;

		case 's':	/* String */
			ctl = gtk_entry_new ();
			gtk_widget_set_size_request(ctl, 400, 0);
			g_settings_bind (settings, k->name, ctl, "text", G_SETTINGS_BIND_DEFAULT);
			break;

		case 'r':	/* Range */
			ctl = range_selector_new(k, settings);
			break;

		default:
			g_warning ("livewallpaper-config : Unrecognized data-type %c for key %s", k->type, k->name);
			break;
	}

	return ctl;
}

static void
add_key(Key *k, GtkWidget *widget, GSettings *settings, GtkWidget *tab)
{
	GtkWidget *ctl;

	/* Create a GtkBox which will contain the key's label and widget */
	GtkWidget *tmpBox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
	gtk_widget_set_tooltip_text (tmpBox, _(k->description));
	gtk_box_pack_start (GTK_BOX(tmpBox),
	                    gtk_label_new (_(k->summary)), FALSE, FALSE, 0);

	/* Reset button */
	ctl = gtk_button_new_from_stock (GTK_STOCK_CLEAR);
	gtk_widget_set_tooltip_text (ctl, _("Reset key to the default value"));
	/* Disable focus */
	gtk_button_set_focus_on_click (GTK_BUTTON(ctl), FALSE);
	/* Join the key name to the widget and bind the signal "clicked" */
	g_object_set_data(G_OBJECT(ctl), "key", g_strdup(k->name));
	g_signal_connect (ctl, "clicked", (GCallback)reset_key, settings);

	gtk_box_pack_end (GTK_BOX(tmpBox), ctl, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(tab), tmpBox, FALSE, FALSE, 0);

	gtk_box_pack_end(GTK_BOX(tmpBox), widget, FALSE, FALSE, 8);
}

GtkWidget*
parse_schema (const gchar *schemafile, const gchar *enumfile, GSettings *settings)
{
	GtkWidget *nb, *tab = NULL; /* The tab which will countain the keys. Reloaded for each tab */
	gchar str[1024];
	Key k;

	FILE *file = fopen(schemafile, "r");
	if(!file)
	{
		perror(schemafile);
		return NULL;
	}

	/* Set locale to C to parse double values locale-independent */
	setlocale(LC_NUMERIC, "C");

	/* The per-plugin notebook */
	nb = gtk_notebook_new();
	gtk_notebook_set_scrollable (GTK_NOTEBOOK(nb), TRUE);

	while (fgets(str, 1024, file))
	{
		if (strstr(str, "<lwp:tab"))
		{
			/* Create a scrolled area for the keys and add it in a new tab */
			GtkWidget *scroll = gtk_scrolled_window_new (NULL, NULL);

			gchar tab_name[32];
			sscanf(strpbrk(str,"n"), "name=" STR_ATTR(31), tab_name);

			tab = gtk_box_new (GTK_ORIENTATION_VERTICAL, 4);
			g_object_set(G_OBJECT(tab),
			             "margin-top", 6,
			             "margin-bottom", 6,
			             "margin-left", 6,
			             "margin-right", 6,
			             NULL);
			gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
			gtk_scrolled_window_add_with_viewport ( GTK_SCROLLED_WINDOW (scroll), tab);
			gtk_notebook_append_page (GTK_NOTEBOOK(nb), scroll, gtk_label_new(_(tab_name)));
		}

		else if (strstr(str, "<lwp:separator"))
		{
			/* Add an horizontal separator to separate widgets inside a same tab */
			GtkWidget *separator = gtk_separator_new(GTK_ORIENTATION_HORIZONTAL);
			g_object_set(G_OBJECT(separator),
			             "margin-left", 6,
			             "margin-right", 6,
			             NULL);
			gtk_box_pack_start (GTK_BOX(tab), separator, FALSE, FALSE, 2);
		}

		else if (strstr(str,"<key "))
		{
			if (strstr(str,"type="))
			{
				/* Get the type and the name of the new key,
				 * and set the number of digits and the incrementation
				 * to a default value */
				if ( sscanf(strpbrk(str,"t"), "type=%*c%c%*c" SPACE "name=" STR_ATTR(31), &k.type, k.name) != 2)
					 sscanf(strpbrk(str,"n"), "name=" STR_ATTR(31) SPACE "type=%*c%c%*c", k.name, &k.type);
				k.digits = 6;
				k.inc    = -1;
			}
			else if (strstr(str, "enum="))
			{
				/* Get the name of an enum key and the associated enum,
				 * and set type to 'e' (not reserved by glib types and
				 * permit to know the key needs load_enum instead of add_key) */
				if ( sscanf(strpbrk(str,"n"), "name=" STR_ATTR(31)  SPACE "enum=" STR_ATTR(127), k.name, k.enum_id) != 2)
					 sscanf(strpbrk(str,"e"), "enum=" STR_ATTR(127) SPACE "name=" STR_ATTR(31),  k.enum_id, k.name);
				k.type = 'e';
			}

			k.skip = FALSE;
		}
		else if (strstr(str,"<lwp:type"))
		{
			gchar buffer[16];
			sscanf(strpbrk(str,">"), STR_TEXT(15), buffer);

			if      (! strcmp(buffer, "image"))   k.type = 'p';
			else if (! strcmp(buffer, "images"))  k.type = 'P';
			else if (! strcmp(buffer, "color"))   k.type = 'c';
			else if (! strcmp(buffer, "font"))    k.type = 'f';
			else if (! strcmp(buffer, "folder"))  k.type = 'F';
			else if (! strcmp(buffer, "folders")) k.type = 'D';
			else if (! strcmp(buffer, "range"))   k.type = 'r';
		}

		else if (strstr(str,"<lwp:skip"))
			k.skip = TRUE;

		else if (strstr(str,"<range"))
			sscanf(strpbrk(str,"m"), "min=%*c%f%*c" SPACE "max=%*c%f%*c", &k.min, &k.max);

		else if (strstr(str,"<summary>"))
			sscanf(strpbrk(str,">"), STR_TEXT(255), k.summary);

		else if (strstr(str,"<description>"))
			sscanf(strpbrk(str,">"), STR_TEXT(511), k.description);

		else if (strstr(str, "<lwp:increment"))
			/* The increment step of the buttons + and - when we do a button1 click 
			 * button2 click = 10*button1 click */
			sscanf(strpbrk(str,">"), ">%f", &k.inc);

		else if (strstr(str, "<lwp:digits"))
			/* The number of decimals displayed */
			sscanf(strpbrk(str,">"), ">%u", &k.digits);

		else if (strstr(str,"</key>") && !k.skip)
		{
			GtkWidget *ctl = NULL;

			if (k.type == 'e')
			{
				ctl = load_enum (&k, schemafile, settings);
				if(ctl == NULL)
					ctl = load_enum (&k, enumfile, settings);
			}
			else
			{
				if (k.type == 'd' || k.type == 'r')
				{
					if (k.inc < 0) k.inc = 10e-7;

					if (nb_decimals (k.inc) > k.digits)
						k.inc = pow (10.0, - (gdouble)k.digits);
				}
				else if (k.inc < 0)
					k.inc = 1;

				ctl = create_widget(&k, settings);
			}

			if(ctl == NULL)
				g_warning ("livewallpaper-config : Could not create widget for %s", k.name);
			else
				add_key (&k, ctl, settings, tab);
		}
	}

	setlocale(LC_NUMERIC, "");

	fclose (file);
	return nb;
}

static void
reset_key (GtkWidget *w, GSettings *s)
{
	/* Reset the key joined to the GtkButton w in GSettings */
	g_settings_reset (s, g_object_get_data(G_OBJECT(w), "key"));
}

static void
cb_bind_get_value (GSettings *s, gchar *key, GtkWidget *w)
{
	/* Get the identifier of the activated choice of the key in GSettings
	 * and set the active choice to this identifier in the GtkComboBoxText */
	gtk_combo_box_set_active_id (GTK_COMBO_BOX(w), g_settings_get_string (s, key));
}

static void
cb_bind_set_value (GtkWidget *w, GSettings *s)
{
	/* Set the current value of the GtkComboBoxText to GSettings
	 * by the identifier of the value */
	g_settings_set_string (s, g_object_get_data(G_OBJECT(w), "key"),
	                       gtk_combo_box_get_active_id (GTK_COMBO_BOX(w)));
}

/* Inspirated by gnome-terminal (profile-editor.c ) */
static gboolean
s_to_rgba (GValue *value, GVariant *variant, gpointer p)
{
	const gchar *s;
	GdkRGBA color;
	(void) p;

	g_variant_get (variant, "&s", &s);
	gdk_rgba_parse (&color, s);

	g_value_set_boxed (value, &color);
	return TRUE;
}

/* Inspirated by gnome-terminal (profile-editor.c ) */
static GVariant *
rgba_to_s (const GValue *v, const GVariantType *t, gpointer p)
{
	(void) t; (void) p;
	return g_variant_new_string (gdk_rgba_to_string (g_value_get_boxed (v)));
}

static void
file_selector_set_value (GtkWidget *w, GSettings *s)
{
	gchar *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (w));
	if (filename) g_settings_set_string (s, g_object_get_data (G_OBJECT (w), "key"), filename);
	g_free (filename);
}

static void
file_selector_get_value (GSettings *s, gchar *key, GtkWidget *w)
{
	gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (w), g_settings_get_string (s, key));
}

