/*
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 * Copyright (C) 2012-2013 RIVIÈRE Aurélien <aurelien.riv@gmail.com>
 *
 */

#include "main.h"


#define KEY_FILE_HAS(p, key) g_key_file_has_key((p)->priv->key_file, "Plugin", key, NULL)
#define KEY_FILE_GET(p, key) g_key_file_get_string((p)->priv->key_file, "Plugin", key, NULL)

#define LW_CONFIG_PLUGIN_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_CONFIG_TYPE_PLUGIN, LwConfigPluginPrivate))

struct _LwConfigPluginPrivate
{
	GKeyFile *key_file;
	gchar *id;
	gchar *name;
	gchar *description;

	gchar *module_dir;
	gchar *data_dir;

	gchar *schema_id;
	GSettings *settings;

	gchar *schema_file;
	gchar *enum_file;

	GtkWidget *settings_widget;
};

G_DEFINE_TYPE(LwConfigPlugin, lw_config_plugin, G_TYPE_OBJECT)

static gchar*
lw_config_plugin_find_schema_file(gchar *filename)
{
	const gchar * const *dirs;
	const gchar *local_dir;
	gchar *path;
	gint i = 0;

	dirs = g_get_system_data_dirs();
	while(dirs[i])
	{
		path = g_build_filename(dirs[i], "glib-2.0", "schemas", filename, NULL);
		if(g_file_test(path, G_FILE_TEST_IS_REGULAR))
			return path;

		g_free(path);
		i++;
	}

	local_dir = g_getenv("GSETTINGS_SCHEMA_DIR");
	if(local_dir != NULL)
	{
		path = g_build_filename(local_dir, filename, NULL);
		if(g_file_test(path, G_FILE_TEST_IS_REGULAR))
			return path;

		g_free(path);
	}

	return NULL;
}

LwConfigPlugin*
lw_config_plugin_new(const gchar *filename,
                     const gchar *module_dir,
                     const gchar *data_dir)
{
	GError *error = NULL;
	LwConfigPlugin *p = g_object_new(LW_CONFIG_TYPE_PLUGIN, NULL);
	GKeyFile *key_file = g_key_file_new();
	gchar *tmp;

	g_key_file_load_from_file(key_file, filename, G_KEY_FILE_NONE, &error);
	if(error != NULL)
	{
		g_critical("Could not load key file: %s", error->message);
		g_error_free(error);
		g_object_unref(p);
		return NULL;
	}

	p->priv->key_file = key_file;
	p->priv->id = KEY_FILE_GET(p, "Module");

	if(p->priv->id == NULL)
	{
		g_critical("Could not get plugin's id from %s", filename);
		g_object_unref(p);
		return NULL;
	}

	p->priv->name = KEY_FILE_GET(p, "Name");
	p->priv->description = KEY_FILE_GET(p, "Description");

	p->priv->module_dir = g_strdup(module_dir);
	p->priv->data_dir = g_strdup(data_dir);

	if(KEY_FILE_HAS(p, "Schema-id"))
		p->priv->schema_id = KEY_FILE_GET(p, "Schema-id");
	else
		p->priv->schema_id = g_strconcat("net.launchpad.livewallpaper.plugins.", p->priv->id, NULL);

	/* Search schema and enum file */
	tmp = g_strconcat(p->priv->schema_id, ".gschema.xml", NULL);
	p->priv->schema_file = lw_config_plugin_find_schema_file(tmp);
	g_free(tmp);

	tmp = g_strconcat(p->priv->schema_id, ".enums.xml", NULL);
	p->priv->enum_file = lw_config_plugin_find_schema_file(tmp);
	g_free(tmp);

	return p;
}

const gchar*
lw_config_plugin_get_id(LwConfigPlugin *self)
{
	return self->priv->id;
}

GdkPixbuf*
lw_config_plugin_get_icon(LwConfigPlugin *self, guint size)
{
	GdkPixbuf *icon;
	gchar *icon_name;

	if(strcmp(self->priv->id, "general") == 0)
		icon_name = g_strdup("livewallpaper");
	else
		icon_name = g_strconcat("livewallpaper-", self->priv->id, NULL);

	icon = gtk_icon_theme_load_icon(gtk_icon_theme_get_default(),
	                                icon_name, size,
	                                GTK_ICON_LOOKUP_FORCE_SIZE, NULL);

	if(icon == NULL)
	{
		icon = gtk_icon_theme_load_icon(gtk_icon_theme_get_default(),
		                                GTK_STOCK_MISSING_IMAGE, size,
		                                GTK_ICON_LOOKUP_FORCE_SIZE, NULL);
	}

	g_free(icon_name);
	return icon;
}

const gchar*
lw_config_plugin_get_name(LwConfigPlugin *self)
{
	return self->priv->name;
}

const gchar*
lw_config_plugin_get_description(LwConfigPlugin *self)
{
	return _(self->priv->description);
}

GSettings*
lw_config_plugin_get_settings(LwConfigPlugin *self)
{
	if(self->priv->settings == NULL)
		self->priv->settings = g_settings_new(self->priv->schema_id);

	g_object_ref(self->priv->settings);
	return self->priv->settings;
}

const gchar*
lw_config_plugin_get_schema_id(LwConfigPlugin *self)
{
	return self->priv->schema_id;
}

const gchar*
lw_config_plugin_get_module_dir(LwConfigPlugin *self)
{
	return self->priv->module_dir;
}

const gchar*
lw_config_plugin_get_data_dir(LwConfigPlugin *self)
{
	return self->priv->data_dir;
}

const gchar*
lw_config_plugin_get_schema_file(LwConfigPlugin *self)
{
	return self->priv->schema_file;
}

const gchar*
lw_config_plugin_get_enum_file(LwConfigPlugin *self)
{
	return self->priv->enum_file;
}

static GtkLicense
get_gtk_license_by_name (gchar *name)
{
	if(! strcmp(name, "GPL2.0"))   return GTK_LICENSE_GPL_2_0;
	if(! strcmp(name, "GPL3.0"))   return GTK_LICENSE_GPL_3_0;
	if(! strcmp(name, "LGPL2.1"))  return GTK_LICENSE_LGPL_2_1;
	if(! strcmp(name, "LGPL3.0"))  return GTK_LICENSE_LGPL_3_0;
	if(! strcmp(name, "BSD"))      return GTK_LICENSE_BSD;
	if(! strcmp(name, "MIT/X11"))  return GTK_LICENSE_MIT_X11;
	if(! strcmp(name, "Artistic")) return GTK_LICENSE_ARTISTIC;

	return GTK_LICENSE_UNKNOWN;
}

GtkAboutDialog*
lw_config_plugin_create_about_dialog(LwConfigPlugin *self)
{
	GtkAboutDialog *dialog = GTK_ABOUT_DIALOG(gtk_about_dialog_new());
	gchar *tmp, **tmpv;

	gtk_about_dialog_set_program_name   (dialog, self->priv->name);
	gtk_about_dialog_set_comments       (dialog, _(self->priv->description));

	if(strcmp(self->priv->id, "general") == 0)
		tmp = g_strdup("livewallpaper");
	else
		tmp = g_strconcat("livewallpaper-", self->priv->id, NULL);
	gtk_about_dialog_set_logo_icon_name (dialog, tmp);
	g_free(tmp);
	tmpv = g_key_file_get_string_list(self->priv->key_file, "Plugin", "Authors", NULL, NULL);
	gtk_about_dialog_set_authors        (dialog, (const gchar**)tmpv);
	g_strfreev(tmpv);
	gtk_about_dialog_set_copyright      (dialog, tmp = KEY_FILE_GET(self, "Copyright"));
	g_free(tmp);
	gtk_about_dialog_set_wrap_license   (dialog, TRUE);

	if(KEY_FILE_HAS(self, "License-type"))
	{
		tmp = KEY_FILE_GET(self, "License-type");
		gtk_about_dialog_set_license_type(dialog, get_gtk_license_by_name(tmp));
		g_free(tmp);
	}
	if(KEY_FILE_HAS(self, "License"))
	{
		gtk_about_dialog_set_license(dialog, tmp = KEY_FILE_GET(self, "License"));
		g_free(tmp);
	}
	if(KEY_FILE_HAS(self, "Website"))
	{
		gtk_about_dialog_set_website(dialog, tmp = KEY_FILE_GET(self, "Website"));
		g_free(tmp);
	}

	return dialog;
}

GtkWidget*
lw_config_plugin_get_settings_widget(LwConfigPlugin *self)
{
	if(self->priv->settings_widget == NULL)
	{
		if(self->priv->schema_file == NULL)
			self->priv->settings_widget = gtk_label_new(_("Cannot find settings for this plugin."));
		else
			self->priv->settings_widget = parse_schema(self->priv->schema_file,
			                                           self->priv->enum_file,
			                                           lw_config_plugin_get_settings(self));
	}

	g_object_ref(self->priv->settings_widget);

	return self->priv->settings_widget;
}

static void
lw_config_plugin_init(LwConfigPlugin *self)
{
	self->priv = LW_CONFIG_PLUGIN_GET_PRIVATE(self);

	self->priv->id = NULL;
	self->priv->key_file = NULL;
	self->priv->name = NULL;
	self->priv->description = NULL;

	self->priv->module_dir = NULL;
	self->priv->data_dir = NULL;

	self->priv->schema_id = NULL;
	self->priv->settings = NULL;

	self->priv->schema_file = NULL;
	self->priv->enum_file = NULL;

	self->priv->settings_widget = NULL;
}

static void
lw_config_plugin_dispose(GObject *object)
{
	LwConfigPlugin *self = LW_CONFIG_PLUGIN(object);

	if(self->priv->key_file)
	{
		g_key_file_unref(self->priv->key_file);
		self->priv->key_file = NULL;
	}

	if(self->priv->settings)
	{
		g_object_unref(self->priv->settings);
		self->priv->settings = NULL;
	}

	if(self->priv->settings_widget)
	{
		gtk_widget_destroy(self->priv->settings_widget);
		self->priv->settings_widget = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_config_plugin_parent_class)->dispose(object);
}

static void
lw_config_plugin_finalize(GObject *object)
{
	LwConfigPlugin *self = LW_CONFIG_PLUGIN(object);

	g_free(self->priv->id);
	g_free(self->priv->name);
	g_free(self->priv->description);
	g_free(self->priv->module_dir);
	g_free(self->priv->data_dir);
	g_free(self->priv->schema_id);
	g_free(self->priv->schema_file);
	g_free(self->priv->enum_file);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_config_plugin_parent_class)->finalize(object);
}

static void
lw_config_plugin_class_init(LwConfigPluginClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = lw_config_plugin_dispose;
	gobject_class->finalize = lw_config_plugin_finalize;

	g_type_class_add_private(klass, sizeof(LwConfigPluginPrivate));
}

