/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 *
 */

/**
 * SECTION: program
 * @Short_description: OpenGL program wrapper
 *
 * #LwProgram is an easy way to use GLSL shaders in a LiveWallpaper plugin.
 *
 * <note>
 *   <para>
 *     LwProgram is not finished yet. It is possible that parts of this type
 *     will be changed in a future version of LiveWallpaper.
 *   </para>
 * </note>
 *
 * <example>
 *   <title>Using LwProgram</title>
 *   <programlisting>
 * GError *error = NULL;
 * LwProgram *prog = g_object_new(LW_TYPE_PROGRAM, NULL);
 *
 * lw_program_create_and_attach_shader(prog, "/path/to/shader.glsl", GL_VERTEX_SHADER, &error);
 * if(error != NULL)
 * {
 *     g_warning("Could not load vertex shader: %s", error->message);
 *     g_error_free(error);
 *     return;
 * }
 *
 * lw_program_link(prog);
 *
 * ...
 *
 * lw_prog_enable(prog);
 *
 * // The program is enabled for all following drawing operations
 * ...
 *
 * lw_prog_disable(prog);</programlisting>
 * </example>
 */

#include <livewallpaper/core.h>


#define LW_PROGRAM_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_TYPE_PROGRAM, LwProgramPrivate))

struct _LwProgramPrivate
{
	guint name;
};

/**
 * LwProgram:
 *
 * A structure for easier OpenGL program handling.
 *
 * Since: 0.4
 */

G_DEFINE_TYPE(LwProgram, lw_program, G_TYPE_OBJECT)

/**
 * lw_program_get_name:
 * @self: A #LwProgram
 *
 * Returns: The program name returned by <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glCreateProgram.xml">glCreateProgram</ulink>
 *
 * Since: 0.4
 */
guint
lw_program_get_name(LwProgram *self)
{
	return self->priv->name;
}

/**
 * lw_program_attach_shader:
 * @self: A #LwProgram
 * @shader: A #LwShader
 *
 * Attaches the shader to the program. This function uses <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glAttachShader.xml">glAttachShader</ulink>
 * to attach the shader. The #LwShader can be freed after attaching it to a #LwProgram.
 *
 * Since: 0.4
 */
void
lw_program_attach_shader(LwProgram *self, LwShader *shader)
{
	glAttachShader(self->priv->name, lw_shader_get_name(shader));
}

/**
 * lw_program_link:
 * @self: A #LwProgram
 *
 * Links the program using <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glLinkProgram.xml">glLinkProgram</ulink>.
 * If an error occurs, this function returns %FALSE and prints a warning.
 *
 * Returns: %TRUE on success, %FALSE if an error occured
 *
 * Since: 0.4
 */
gboolean
lw_program_link(LwProgram *self)
{
	int status;

	glLinkProgram(self->priv->name);

	/* Handle errors */
	glGetProgramiv(self->priv->name, GL_LINK_STATUS, &status);
	if(status == GL_FALSE)
	{
		int log_length;
		gchar *log_buffer;

		glGetProgramiv(self->priv->name, GL_INFO_LOG_LENGTH, &log_length);
		log_buffer = g_malloc(log_length * sizeof(gchar));

		glGetProgramInfoLog(self->priv->name, log_length, NULL, log_buffer);

		g_warning("%s", log_buffer);

		g_free(log_buffer);
		return FALSE;
	}

	return TRUE;
}

/**
 * lw_program_create_and_attach_shader:
 * @self: A #LwProgram
 * @filename: The file containing the shader's source code
 * @type: GL_VERTEX_SHADER or GL_FRAGMENT_SHADER
 * @error: return location for a #GError, or %NULL
 *
 * This function creates, compiles and attaches a shader to the program.
 * It is easier to use this function instead of creating, compiling and
 * attaching the shader by yourself.
 *
 * Returns: %TRUE on success, %FALSE otherwise
 *
 * Since: 0.4
 */
gboolean
lw_program_create_and_attach_shader(LwProgram *self, const gchar *filename, guint type, GError **error)
{
	LwShader *shader = lw_shader_new_from_file(filename, type, error);

	if(*error == NULL)
	{
		lw_shader_compile(shader);
		lw_program_attach_shader(self, shader);
		g_object_unref(shader);

		return TRUE;
	}

	return FALSE;
}

/**
 * lw_program_get_attrib_location:
 * @self: A #LwProgram
 * @name: Name of an attribute variable
 *
 * Calls <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml">glGetAttribLocation</ulink>
 * to get the location of an attribute variable.
 *
 * Returns: the location of an attribute variable
 *
 * Since: 0.4
 */
gint
lw_program_get_attrib_location(LwProgram *self, const gchar *name)
{
	return glGetAttribLocation(self->priv->name, name);
}

/**
 * lw_program_get_uniform_location:
 * @self: A #LwProgram
 * @name: Name of an uniform variable
 *
 * Calls <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml">glGetUniformLocation</ulink>
 * to get the location of an uniform variable.
 *
 * Returns: the location of an uniform variable
 *
 * Since: 0.4
 */
gint
lw_program_get_uniform_location(LwProgram *self, const gchar *name)
{
	return glGetUniformLocation(self->priv->name, name);
}

/**
 * lw_program_enable:
 * @self: A #LwProgram
 *
 * Enables the program.
 *
 * Since: 0.4
 */
void
lw_program_enable(LwProgram *self)
{
	glUseProgram(self->priv->name);
}

/**
 * lw_program_disable:
 * @self: A #LwProgram
 *
 * Disables this program.
 *
 * Since: 0.4
 */
void
lw_program_disable(LwProgram *self)
{
	(void) self;

	glUseProgram(0);
}

static void
lw_program_init(LwProgram *self)
{
	self->priv = LW_PROGRAM_GET_PRIVATE(self);

	self->priv->name = glCreateProgram();
}

static void
lw_program_finalize(GObject *object)
{
	LwProgram *self = LW_PROGRAM(object);

	if(self->priv->name)
		glDeleteProgram(self->priv->name);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_program_parent_class)->finalize(object);
}

static void
lw_program_class_init(LwProgramClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = lw_program_finalize;

	g_type_class_add_private(klass, sizeof(LwProgramPrivate));
}

