/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012-2013 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 *
 */

/**
 * SECTION: util
 * @Short_description: functions to simplify some common tasks
 * @Title: Miscellaneous Utilities
 *
 * This section provides several functions and macros that reduce the code
 * size or simplify some common tasks.
 */

#include <livewallpaper/core.h>

/**
 * LW_UNREF_GOBJECT:
 * @object: A #GObject that needs to be freed with g_object_unref()
 *
 * Frees the @object with g_object_unref() and sets it to %NULL, but only
 * if the object is not %NULL. This macro is mostly used in the dispose method
 * of a #GObject if the object holds references of other objects. The only
 * advantage of this macro is to make the source files a little bit cleaner.
 *
 * Since: 0.4
 */

/**
 * lw_settings_bind_color:
 * @settings: A #GSettings object
 * @key: The key to bind
 * @object: (type GObject.Object): A #GObject
 * @property: The name of the property to bind
 * @flags: Flags for the binding
 *
 * This function calls g_settings_bind_with_mapping() with the required
 * mapping functions for colors. It uses lw_color_get_color_mapping()
 * and lw_color_set_color_mapping() as mapping functions.
 *
 * Since: 0.4
 */
void
lw_settings_bind_color(GSettings *settings,
                       const gchar *key,
                       gpointer object,
                       const gchar *property,
                       GSettingsBindFlags flags)
{
	g_settings_bind_with_mapping(settings, key, object, property, flags,
		lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
}

/**
 * lw_settings_bind_range:
 * @settings: A #GSettings object
 * @key: The key to bind
 * @object: (type GObject.Object): A #GObject
 * @property: The name of the property to bind
 * @flags: Flags for the binding
 *
 * This function calls g_settings_bind_with_mapping() with the required
 * mapping functions for #LwRange properties. It uses lw_range_get_range_mapping()
 * and lw_range_set_range_mapping() as mapping functions.
 *
 * Since: 0.4
 */
void
lw_settings_bind_range(GSettings *settings,
                       const gchar *key,
                       gpointer object,
                       const gchar *property,
                       GSettingsBindFlags flags)
{
	g_settings_bind_with_mapping(settings, key, object, property, flags,
		lw_range_get_range_mapping, lw_range_set_range_mapping, NULL, NULL);
}

typedef struct
{
	GSettings *settings;
	gchar *key;
} LwEnumBinding;

static gboolean
lw_enum_get_enum_mapping(GValue *value, GVariant *variant, LwEnumBinding *binding)
{
	(void) variant;

	g_value_set_uint(value, g_settings_get_enum(binding->settings, binding->key));
	return TRUE;
}

static GVariant*
lw_enum_set_enum_mapping(const GValue *value, const GVariantType *type, LwEnumBinding *binding)
{
	(void) type;

	/*
	 * We set the enum here, even if it is not the purpose of this function.
	 * The disadvantage of this is, that it also calls lw_enum_get_enum_mapping.
	 */
	g_settings_set_enum(binding->settings, binding->key, g_value_get_uint(value));
	return g_variant_new_string(g_settings_get_string(binding->settings, binding->key));
}

static void
lw_enum_binding_free(LwEnumBinding *binding)
{
	g_free(binding->key);
	g_slice_free(LwEnumBinding, binding);
}

/**
 * lw_settings_bind_enum:
 * @settings: A #GSettings object
 * @key: The key to bind
 * @object: (type GObject.Object): A #GObject
 * @property: The name of the property to bind
 * @flags: Flags for the binding
 *
 * Creates a binding between an enum @key and a #guint @property. This function
 * passes all params to g_settings_bind_with_mapping() and sets the required
 * mapping functions.
 *
 * Since: 0.4
 */
void
lw_settings_bind_enum(GSettings *settings,
                      const gchar *key,
                      gpointer object,
                      const gchar *property,
                      GSettingsBindFlags flags)
{
	LwEnumBinding *binding = g_slice_new(LwEnumBinding);

	binding->settings = settings;
	binding->key = g_strdup(key);

	g_settings_bind_with_mapping(settings, key, object, property, flags,
	                             (GSettingsBindGetMapping) lw_enum_get_enum_mapping,
	                             (GSettingsBindSetMapping) lw_enum_set_enum_mapping,
	                             binding, (GDestroyNotify) lw_enum_binding_free);
}

